use axum_jwt_login::{AuthBackend, UserPermissions};
use ldap::LDAPBackend;
use sqlx::PgPool;

use crate::{
    api::routes::users::{models::UserStatus, sql::get_users},
    config::Configuration,
    errors::ApiError,
};

use super::routes::{auth::models::Credentials, users::models::User};

pub mod ldap;

#[derive(Debug, Clone)]
pub struct ApiBackend {
    pool: PgPool,
    config: Configuration,
}

impl ApiBackend {
    pub fn new(pool: PgPool, config: &Configuration) -> Self {
        Self {
            pool,
            config: config.clone(),
        }
    }

    pub fn pool(&self) -> &PgPool {
        &self.pool
    }

    pub fn config(&self) -> &Configuration {
        &self.config
    }
}

impl AuthBackend<User> for ApiBackend {
    type Credentials = Credentials;
    type Error = ApiError;

    async fn authenticate(
        &self,
        Credentials { id, password }: Self::Credentials,
    ) -> Result<Option<User>, Self::Error> {
        // get user from Database
        let user = get_users(&self.pool, Some(UserStatus::Active), Some(id)).await?;
        let user = user.first().ok_or(ApiError::InvalidCredentials)?;

        // authenticate user
        if user.active_directory_auth {
            // authenticate against LDAP (AD) server
            // --
            let mut ldap = LDAPBackend::from_config(&self.config).await?;
            ldap.ad_bind(&user.id(), &password).await?;

            // update user information
            let details = ldap.get_user_detail(&user.id()).await?;
            user.update_with_ad_details(&self.pool, details).await?;

            // terminate connection
            ldap.unbind().await;
        } else {
            argon2::verify_encoded(&user.password, password.as_bytes())
                .map_err(|_| ApiError::InvalidCredentials)?;
        }

        Ok(Some(user.clone()))
    }
}
