use axum::{debug_handler, extract::Query, Extension, Json};
use models::ApiKey;
use serde::Deserialize;
use utoipa::IntoParams;
use utoipa_axum::{router::OpenApiRouter, routes};

use crate::{
    api::{
        backend::ApiBackend,
        description::API_KEY_TAG,
        routes::{
            users::permissions::{Permission, PermissionDetail},
            User,
        },
    },
    errors::ApiError,
    permission_required,
};

use super::AuthBackendType;

pub mod models;
pub mod sql;

// expose the OpenAPI to parent module
pub fn router() -> OpenApiRouter {
    let read = OpenApiRouter::new()
        .routes(routes!(get_api_keys))
        .route_layer(permission_required!(Permission::Read(
            PermissionDetail::APIKeys
        )));
    let write = OpenApiRouter::new()
        .routes(routes!(create_api_key, update_api_key, delete_api_key))
        .route_layer(permission_required!(Permission::Write(
            PermissionDetail::APIKeys
        )));

    OpenApiRouter::new().merge(read).merge(write)
}

#[debug_handler]
#[utoipa::path(
    get,
    path = "/apikeys",
    summary = "Get all API Keys",
    description = "Get a list of all configured API Keys.",
    responses(
        (status = OK, body = Vec<ApiKey>, description = "List of API Keys"),
    ),
    security(
        ("user_auth" = ["read:apikeys",]),
    ),
    tag = API_KEY_TAG)]
pub async fn get_api_keys(
    Extension(backend): Extension<ApiBackend>,
) -> Result<Json<Vec<ApiKey>>, ApiError> {
    Ok(Json(sql::get_api_keys(backend.pool()).await?))
}

#[debug_handler]
#[utoipa::path(
    post,
    path = "/apikeys",
    summary = "Create new API Key",
    description = "Create a new API Key.",
    request_body(content = ApiKey, description = "API Key details", content_type = "application/json"),
    responses(
        (status = OK, description = "API Key successfully created (API Key Secret in Body)", body = String),
    ),
    security(
        ("user_auth" = ["write:apikeys",]),
    ),
    tag = API_KEY_TAG)]
pub async fn create_api_key(
    auth_session: AuthBackendType,
    Json(api_key): Json<ApiKey>,
) -> Result<String, ApiError> {
    let backend = auth_session.backend();

    // create new API key
    let (key_secret, key) = ApiKey::create(
        &api_key.name,
        api_key.auth_required,
        api_key.permissions.0,
        backend.config(),
    )?;

    // insert API Key into database
    sql::create_api_key(backend.pool(), &key).await?;

    // add API Key to session
    auth_session.add_api_key(key).await;

    // Return key secret in response
    Ok(key_secret)
}

#[debug_handler]
#[utoipa::path(
    put,
    path = "/apikeys",
    summary = "Update API Key",
    description = "Update an API Key.",
    request_body(content = ApiKey, description = "API Key details", content_type = "application/json"),
    responses(
        (status = OK, description = "API Key successfully updated"),
    ),
    security(
        ("user_auth" = ["write:apikeys",]),
    ),
    tag = API_KEY_TAG)]
pub async fn update_api_key(
    Extension(backend): Extension<ApiBackend>,
    Json(api_key): Json<ApiKey>,
) -> Result<(), ApiError> {
    sql::update_api_key(backend.pool(), &api_key).await?;
    Ok(())
}

#[derive(Debug, Deserialize, IntoParams)]
pub struct DeleteQueryParameters {
    key_id: String,
}

#[debug_handler]
#[utoipa::path(
    delete,
    path = "/apikeys",
    summary = "Delete API Key",
    description = "Delete an API Key.",
    params(DeleteQueryParameters),
    responses(
        (status = OK, description = "API Key successfully deleted"),
    ),
    security(
        ("user_auth" = ["write:apikeys",]),
    ),
    tag = API_KEY_TAG)]
pub async fn delete_api_key(
    Extension(backend): Extension<ApiBackend>,
    Query(params): Query<DeleteQueryParameters>,
) -> Result<(), ApiError> {
    sql::delete_api_key(backend.pool(), &params.key_id).await?;
    Ok(())
}
