use sqlx::{PgPool, Postgres, Transaction};

use crate::errors::ApiError;

use super::models::ApiKey;

pub async fn get_api_keys(pool: &PgPool) -> Result<Vec<ApiKey>, ApiError> {
    Ok(sqlx::query_as!(
        ApiKey,
        r#"SELECT
            APIKEYS."KeyID" as id,
            APIKEYS."Name" as name,
            APIKEYS."Hash" as hash,
            APIKEYS."UserAuthRequired" as auth_required,
            APIKEYS."CreationDate" as "creation_date?",
            APIKEYS."LastChanged" as "last_change?",
            NULL as api_config_secret,
            array_remove(ARRAY_AGG(APIKEY_PERMISSIONS."Permission"), NULL) AS permissions
        FROM
            apikeys
        LEFT JOIN APIKEY_PERMISSIONS ON APIKEY_PERMISSIONS."KeyID" = APIKEYS."KeyID"
        GROUP BY APIKEYS."KeyID""#
    )
    .fetch_all(pool)
    .await?)
}

pub async fn create_api_key(pool: &PgPool, api_key: &ApiKey) -> Result<(), ApiError> {
    // start transaction
    let mut transaction = pool.begin().await?;

    // create api_key
    sqlx::query!(
        r#"INSERT INTO apikeys
        ("KeyID", "Name", "UserAuthRequired", "Hash") 
        VALUES ($1, $2, $3, $4)"#,
        api_key.id,
        api_key.name,
        api_key.auth_required,
        api_key.hash,
    )
    .execute(&mut *transaction)
    .await?;

    // update permissions
    update_permissions(&mut transaction, &api_key).await?;

    // commit transaction
    transaction.commit().await?;

    Ok(())
}

pub async fn update_api_key(pool: &PgPool, api_key: &ApiKey) -> Result<(), ApiError> {
    // start transaction
    let mut transaction = pool.begin().await?;

    // create api_key
    sqlx::query!(
        r#"UPDATE apikeys SET
            "Name" = $2,
            "UserAuthRequired" = $3
        WHERE "KeyID" = $1"#,
        api_key.id,
        api_key.name,
        api_key.auth_required,
    )
    .execute(&mut *transaction)
    .await?;

    // update permissions
    update_permissions(&mut transaction, &api_key).await?;

    // commit transaction
    transaction.commit().await?;

    Ok(())
}

pub async fn update_permissions(
    transaction: &mut Transaction<'static, Postgres>,
    api_key: &ApiKey,
) -> Result<(), ApiError> {
    // delete all permissions for API Key
    sqlx::query!(
        r#"DELETE FROM apikey_permissions
        WHERE "KeyID" = $1"#,
        api_key.id,
    )
    .execute(&mut **transaction)
    .await?;

    // insert permissions
    for permission in &api_key.permissions.0 {
        sqlx::query!(
            r#"INSERT INTO apikey_permissions
            ("KeyID", "Permission") 
            VALUES ($1, $2)"#,
            api_key.id,
            permission.to_string()
        )
        .execute(&mut **transaction)
        .await?;
    }

    Ok(())
}

pub async fn delete_api_key(pool: &PgPool, key_id: &str) -> Result<(), ApiError> {
    // start transaction
    let mut transaction = pool.begin().await?;

    // Delete permissions
    sqlx::query!(
        r#"DELETE FROM apikey_permissions
        WHERE "KeyID" = $1"#,
        key_id,
    )
    .execute(&mut *transaction)
    .await?;

    // Delete API Key
    sqlx::query!(
        r#"DELETE FROM apikeys
        WHERE "KeyID" = $1"#,
        key_id,
    )
    .execute(&mut *transaction)
    .await?;

    // commit transaction
    transaction.commit().await?;

    Ok(())
}
