use std::io::Cursor;

use axum::{debug_handler, Json};
use chrono::Utc;
use minisign::{KeyPair, PublicKeyBox, SecretKeyBox, SignatureBox};
use models::{
    PublicKeyID, SignatureTest, SigningKeyRequest, SigningKeyStatus, VerificationRequest,
};
use utoipa_axum::{router::OpenApiRouter, routes};

use crate::{
    api::{
        description::SIGNATURE_TAG,
        routes::users::permissions::{Permission, PermissionDetail},
    },
    errors::ApiError,
    permission_required,
};

use super::AuthBackendType;

pub mod models;
pub mod sql;

// expose the OpenAPI to parent module
pub fn router() -> OpenApiRouter {
    let verify = OpenApiRouter::new().routes(routes!(verify));
    let write = OpenApiRouter::new()
        .routes(routes!(
            check_signing_key,
            create_signing_key,
            verify_signing_key
        ))
        .routes(routes!(sign))
        .route_layer(permission_required!(Permission::Write(
            PermissionDetail::Signature
        )));

    OpenApiRouter::new().merge(write).merge(verify)
}

#[debug_handler]
#[utoipa::path(
    get,
    path = "/signature/key",
    summary = "Check for signing key",
    description = "Check if the logged in user has configured a signing key.",
    responses(
        (status = OK, body = SigningKeyStatus, description = "Signature Status"),
    ),
    security(
        ("user_auth" = ["write:signature",]),
    ),
    tag = SIGNATURE_TAG)]
pub async fn check_signing_key(
    auth_session: AuthBackendType,
) -> Result<Json<SigningKeyStatus>, ApiError> {
    let backend = auth_session.backend();
    let user = auth_session
        .is_authenticated()
        .ok_or(ApiError::AccessDenied)?;

    let private_key = sql::get_private_key(backend.pool(), &user.user_id).await?;

    Ok(Json(SigningKeyStatus {
        configured: private_key.key.is_some(),
        verified: None,
    }))
}

#[debug_handler]
#[utoipa::path(
    post,
    path = "/signature/key",
    summary = "Create signing key",
    description = "Create signing key for the currently logged in user.",
    request_body(content = SigningKeyRequest, description = "Signing Key Request", content_type = "application/json"),
    responses(
        (status = OK, description = "Successfully created signing key"),
    ),
    security(
        ("user_auth" = ["write:signature",]),
    ),
    tag = SIGNATURE_TAG)]
pub async fn create_signing_key(
    auth_session: AuthBackendType,
    Json(key_request): Json<SigningKeyRequest>,
) -> Result<(), ApiError> {
    let backend = auth_session.backend();
    let user = auth_session
        .is_authenticated()
        .ok_or(ApiError::AccessDenied)?;

    // create Key Pair
    let KeyPair { pk, sk } = KeyPair::generate_encrypted_keypair(Some(key_request.secret))?;
    let public_key = pk.to_box()?.to_string();
    let private_key = sk
        .to_box(Some(&format!(
            "{} {} - {}",
            user.name,
            user.surname,
            Utc::now()
        )))? // Optional comment about the key
        .to_string();

    // store keys in database
    sql::update_private_key(backend.pool(), &user.user_id, private_key).await?;
    sql::store_public_key(backend.pool(), &user.user_id, public_key).await?;

    Ok(())
}

#[debug_handler]
#[utoipa::path(
    put,
    path = "/signature/key",
    summary = "Verify signing key",
    description = "Verify signing key for the currently logged in user.",
    request_body(content = SigningKeyRequest, description = "Signing Key Request", content_type = "application/json"),
    responses(
        (status = OK, body = SigningKeyStatus, description = "Signature Status"),
    ),
    security(
        ("user_auth" = ["write:signature",]),
    ),
    tag = SIGNATURE_TAG)]
pub async fn verify_signing_key(
    auth_session: AuthBackendType,
    Json(key_request): Json<SigningKeyRequest>,
) -> Result<Json<SigningKeyStatus>, ApiError> {
    let backend = auth_session.backend();
    let user = auth_session
        .is_authenticated()
        .ok_or(ApiError::AccessDenied)?;

    // get private key
    let status = match key_request
        .to_private_key(backend.pool(), &user.user_id)
        .await?
    {
        Some(_key) => SigningKeyStatus {
            configured: true,
            verified: Some(true),
        },
        None => SigningKeyStatus {
            configured: false,
            verified: None,
        },
    };

    Ok(Json(status))
}

#[debug_handler]
#[utoipa::path(
    post,
    path = "/signature/sign",
    summary = "Sign something with your signing key",
    description = "Sign something as the currently logged in user.",
    request_body(content = SigningKeyRequest, description = "Signing Key Request", content_type = "application/json"),
    responses(
        (status = OK, description = "Successfully signed"),
    ),
    security(
        ("user_auth" = ["write:signature",]),
    ),
    tag = SIGNATURE_TAG)]
pub async fn sign(
    auth_session: AuthBackendType,
    Json(key_request): Json<SigningKeyRequest>,
) -> Result<Json<VerificationRequest>, ApiError> {
    let backend = auth_session.backend();
    let user = auth_session
        .is_authenticated()
        .ok_or(ApiError::AccessDenied)?;

    // get private key
    let private_key = key_request
        .to_private_key(backend.pool(), &user.user_id)
        .await?
        .ok_or(ApiError::AccessDenied)?;

    // get current public key
    let public_key_id = sql::get_current_public_key(backend.pool(), &user.user_id)
        .await?
        .ok_or(ApiError::InternalError(
            "No corresponding public key found".to_string(),
        ))?;

    let data = SignatureTest {
        value1: "blfjdljfa".to_string(),
        value2: false,
        value3: vec!["bladfa".to_string(), "noch bla".to_string()],
    };

    let data_reader = Cursor::new(data.calculate_hash().to_be_bytes());
    let signature_box = minisign::sign(
        None,
        &private_key,
        data_reader,
        None,
        Some(&serde_json::to_string(&public_key_id)?),
    )?;

    // We have a signature! Let's inspect it a little bit.
    println!(
        "Untrusted comment: [{}]",
        signature_box.untrusted_comment().unwrap()
    );
    println!(
        "Trusted comment: [{}]",
        signature_box.trusted_comment().unwrap()
    );

    // store signature
    // Converting the signature box to a string in order to save it is easy.
    let signature = signature_box.into_string();
    println!("BLA");
    println!("{signature}");

    Ok(Json(VerificationRequest {
        signature,
        signed_from: "".to_string(),
    }))
}

#[debug_handler]
#[utoipa::path(
    get,
    path = "/signature/verify",
    summary = "Verify a signature",
    description = "Verify the signature.",
    request_body(content = SigningKeyRequest, description = "Signing Key Request", content_type = "application/json"),
    responses(
        (status = OK, description = "Successfully signed"),
    ),
    tag = SIGNATURE_TAG)]
pub async fn verify(
    auth_session: AuthBackendType,
    Json(verification_request): Json<VerificationRequest>,
) -> Result<(), ApiError> {
    let backend = auth_session.backend();

    // Now, let's verify the signature.
    // Assuming we just loaded it into `signature_box_str`, get the box back.
    let signature_box = SignatureBox::from_string(&verification_request.signature)?;

    // get public key id from signature
    let public_key_id: PublicKeyID = serde_json::from_str(&signature_box.untrusted_comment()?)?;

    // Load the public key of the user that signed.
    let public_key_str = sql::get_public_key_by_id(backend.pool(), public_key_id)
        .await?
        .key
        .ok_or(ApiError::InternalError(
            "No corresponding Public Key was found.".to_string(),
        ))?;
    let public_key_box = PublicKeyBox::from_string(&public_key_str)?;
    let public_key = public_key_box.into_public_key()?;

    // And verify the data.
    let data = SignatureTest {
        value1: "blfjdljfa".to_string(),
        value2: false,
        value3: vec!["bladfa".to_string(), "noch bla".to_string()],
    };
    let data_reader = Cursor::new(data.calculate_hash().to_be_bytes());
    let verified = minisign::verify(&public_key, &signature_box, data_reader, true, false, false);
    match verified {
        Ok(()) => println!("Success!"),
        Err(_) => println!("Verification failed"),
    };

    Ok(())
}
