use std::collections::{HashMap, HashSet};

use axum_jwt_login::UserPermissions;
use chrono::NaiveDateTime;
use serde::{Deserialize, Serialize};
use sqlx::PgPool;
use ts_rs::TS;
use utoipa::ToSchema;

use crate::errors::ApiError;

use super::{
    permissions::{Permission, PermissionContainer},
    sql,
};

#[derive(Debug, Clone, Copy, Serialize, Deserialize, sqlx::Type, TS, ToSchema)]
#[repr(i16)]
pub enum UserStatus {
    Deleted = -1,
    Deactivated = 0,
    Active = 1,
}

#[derive(Debug, Clone, Serialize, Deserialize, TS, ToSchema)]
pub struct GroupContainer(pub HashSet<i32>);

impl From<Option<Vec<i32>>> for GroupContainer {
    fn from(value: Option<Vec<i32>>) -> Self {
        let set = match value {
            Some(values) => HashSet::from_iter(values),
            None => HashSet::new(),
        };

        Self(set)
    }
}

#[derive(Debug, Clone, Serialize, Deserialize, TS, ToSchema)]
#[ts(export)]
pub struct User {
    pub user_id: String,
    pub active_directory_auth: bool,
    #[serde(skip)] // do not leak password hashes
    pub password: String,
    pub name: String,
    pub surname: String,
    pub email: String,
    #[schema(inline)]
    pub groups: GroupContainer,
    #[schema(inline)]
    pub group_permissions: PermissionContainer,
    #[schema(inline)]
    pub permissions: PermissionContainer,
    pub status_flag: UserStatus,
    #[serde(default)]
    #[schema(value_type = String, read_only)]
    pub creation_date: NaiveDateTime,
    #[serde(default)]
    #[schema(value_type = String, read_only)]
    pub last_change: NaiveDateTime,
}

impl UserPermissions for User {
    type Error = ApiError;
    type Permission = Permission;
    type Id = String;

    fn id(&self) -> Self::Id {
        self.user_id.clone()
    }

    fn get_user_permissions(&self) -> Result<HashSet<Self::Permission>, Self::Error> {
        Ok(self.permissions.0.iter().cloned().collect())
    }
    fn get_group_permissions(&self) -> Result<HashSet<Self::Permission>, Self::Error> {
        Ok(self.group_permissions.0.iter().cloned().collect())
    }
}

impl User {
    pub async fn update_with_ad_details(
        &self,
        pool: &PgPool,
        details: HashMap<String, Vec<String>>,
    ) -> Result<(), ApiError> {
        let create_error = |key: &str| {
            ApiError::InternalError(format!(
                "Update User Information: Information '{key}' is not present"
            ))
        };

        // get field values
        let name = details
            .get("givenName")
            .ok_or(create_error("Name"))?
            .first()
            .ok_or(create_error("Name"))?;
        let surname = details
            .get("sn")
            .ok_or(create_error("Surname"))?
            .first()
            .ok_or(create_error("Surname"))?;
        let email = details
            .get("mail")
            .ok_or(create_error("Email"))?
            .first()
            .ok_or(create_error("Email"))?;

        if name != &self.name || surname != &self.surname || email != &self.email {
            let mut transaction = pool.begin().await?;
            sql::update_user_details(&mut transaction, &self.id(), name, surname, email).await?;
            transaction.commit().await?;
        }

        Ok(())
    }
}
