use std::{collections::HashSet, convert::Infallible, fmt::Display, str::FromStr};

use serde::{Deserialize, Serialize};
use strum::EnumString;
use ts_rs::TS;
use utoipa::ToSchema;

#[derive(Debug, Clone, Copy, Serialize, Deserialize, PartialEq, Eq, Hash, TS, ToSchema)]
pub enum Permission {
    Read(PermissionDetail),
    Write(PermissionDetail),
    None,
}

impl FromStr for Permission {
    type Err = Infallible;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        if let Some((permission, detail)) = s.split_once(':') {
            match permission {
                "Read" => Ok(Self::Read(
                    PermissionDetail::from_str(detail).unwrap_or_default(),
                )),
                "Write" => Ok(Self::Write(
                    PermissionDetail::from_str(detail).unwrap_or_default(),
                )),
                _ => Ok(Self::None),
            }
        } else {
            Ok(Self::None)
        }
    }
}

impl Display for Permission {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let string = match self {
            Permission::Read(detail) => format!("Read:{detail}"),
            Permission::Write(detail) => format!("Write:{detail}"),
            Permission::None => "None".to_string(),
        };
        f.write_str(&string)
    }
}

#[derive(
    Debug,
    Default,
    Clone,
    Copy,
    Serialize,
    Deserialize,
    PartialEq,
    Eq,
    Hash,
    EnumString,
    strum::Display,
    TS,
    ToSchema,
)]
pub enum PermissionDetail {
    Users,
    APIKeys,
    Signature,
    #[default]
    None,
}

#[derive(Debug, Clone, Serialize, Deserialize, TS, ToSchema)]
pub struct PermissionContainer(pub HashSet<Permission>);

impl From<Option<Vec<String>>> for PermissionContainer {
    fn from(value: Option<Vec<String>>) -> Self {
        let set = match value {
            Some(values) => HashSet::from_iter(
                values
                    .iter()
                    .map(|s| Permission::from_str(s).unwrap_or(Permission::None)),
            ),
            None => HashSet::new(),
        };

        Self(set)
    }
}

#[cfg(test)]
mod test {
    use std::str::FromStr;

    use crate::api::routes::users::permissions::PermissionDetail;

    use super::Permission;

    #[tokio::test]
    async fn test_permissions() {
        let permission = Permission::Read(PermissionDetail::Users);
        assert_eq!("Read:Users", permission.to_string());

        let parsed = Permission::from_str("Write:Users").unwrap();
        assert_eq!(parsed, Permission::Write(PermissionDetail::Users));
    }
}
