use sqlx::{PgPool, Postgres, Transaction};

use crate::errors::ApiError;

use super::models::{User, UserStatus};

pub async fn get_users(
    pool: &PgPool,
    filter_status: Option<UserStatus>,
    filter_id: Option<String>,
) -> Result<Vec<User>, ApiError> {
    Ok(sqlx::query_as!(
        User,
        r#"SELECT
            USERS."UserID" as user_id,
            USERS."ActiveDirectoryAuth" as active_directory_auth,
            USERS."Password" as password,
            USERS."Name" as name,
            USERS."Surname" as surname,
            USERS."Email" as email,
            USERS."StatusFlag" as "status_flag: UserStatus",
            USERS."CreationDate" as "creation_date",
            USERS."LastChanged" as "last_change",
            array_remove(ARRAY_AGG(USERS_GROUPS."GroupID"), NULL) AS groups,
            array_remove(ARRAY_AGG(USER_PERMISSIONS."Permission"), NULL) AS permissions,
            array_remove(ARRAY_AGG(GROUP_PERMISSIONS."Permission"), NULL) AS group_permissions
        FROM
            users
        LEFT JOIN PUBLIC.USER_PERMISSIONS ON USER_PERMISSIONS."UserID" = USERS."UserID"
        LEFT JOIN USERS_GROUPS ON USERS."UserID" = USERS_GROUPS."UserID"
        LEFT JOIN GROUP_PERMISSIONS ON GROUP_PERMISSIONS."GroupID" = USERS_GROUPS."GroupID"
        WHERE
            ($1::smallint IS NULL OR USERS."StatusFlag" = $1)
            AND ($2::varchar IS NULL OR USERS."UserID" = $2)
        GROUP BY USERS."UserID""#,
        filter_status.map(|s| s as i16),
        filter_id
    )
    .fetch_all(pool)
    .await?)
}

pub async fn update_user(pool: &PgPool, user: &User) -> Result<(), ApiError> {
    let mut transaction = pool.begin().await?;

    // update general stuff
    sqlx::query!(
        r#"UPDATE users SET
            "ActiveDirectoryAuth" = $2,
            "StatusFlag" = $3,
            "LastChanged" = NOW()
        WHERE "UserID" = $1"#,
        user.user_id,
        user.active_directory_auth,
        user.status_flag as i16,
    )
    .execute(&mut *transaction)
    .await?;

    // update user information
    update_user_details(
        &mut transaction,
        &user.user_id,
        &user.name,
        &user.surname,
        &user.email,
    )
    .await?;
    // update permissions
    update_user_permissions(&mut transaction, &user).await?;
    // update groups
    update_user_groups(&mut transaction, user).await?;

    // commit transaction
    transaction.commit().await?;

    Ok(())
}

pub async fn update_user_details<'a>(
    pool: &'a mut Transaction<'static, Postgres>,
    user_id: &String,
    name: &String,
    surname: &String,
    email: &String,
) -> Result<(), ApiError> {
    // let test = **pool;
    sqlx::query!(
        r#"UPDATE users SET
            "Name" = $2,
            "Surname" = $3,
            "Email" = $4
        WHERE "UserID" = $1"#,
        user_id,
        name,
        surname,
        email
    )
    .execute(&mut **pool)
    .await?;

    Ok(())
}

pub async fn update_user_permissions<'a>(
    pool: &'a mut Transaction<'static, Postgres>,
    user: &User,
) -> Result<(), ApiError> {
    // delete all permissions for user
    sqlx::query!(
        r#"DELETE FROM user_permissions
        WHERE "UserID" = $1"#,
        user.user_id,
    )
    .execute(&mut **pool)
    .await?;

    // insert permissions
    for permission in &user.permissions.0 {
        sqlx::query!(
            r#"INSERT INTO user_permissions
            ("UserID", "Permission") 
            VALUES ($1, $2)"#,
            user.user_id,
            permission.to_string()
        )
        .execute(&mut **pool)
        .await?;
    }

    Ok(())
}

pub async fn update_user_groups<'a>(
    pool: &'a mut Transaction<'static, Postgres>,
    user: &User,
) -> Result<(), ApiError> {
    // delete all permissions for user
    sqlx::query!(
        r#"DELETE FROM users_groups
        WHERE "UserID" = $1"#,
        user.user_id,
    )
    .execute(&mut **pool)
    .await?;

    // insert permissions
    for group_id in &user.groups.0 {
        sqlx::query!(
            r#"INSERT INTO users_groups
            ("UserID", "GroupID") 
            VALUES ($1, $2)"#,
            user.user_id,
            group_id
        )
        .execute(&mut **pool)
        .await?;
    }

    Ok(())
}

pub async fn create_new_user(
    pool: &PgPool,
    user: &User,
    password: Option<String>,
) -> Result<(), ApiError> {
    let mut transaction = pool.begin().await?;

    sqlx::query!(
        r#"INSERT INTO users
        ("UserID", "Password") 
        VALUES ($1, $2)"#,
        user.user_id,
        password
    )
    .execute(&mut *transaction)
    .await?;

    // commit transaction
    transaction.commit().await?;

    Ok(())
}
