use rand::{distributions::Alphanumeric, Rng};
use sha2::{Digest, Sha512};
use uuid::Uuid;

pub(crate) const KEY_LENGTH: usize = 40;

pub trait ApiKeyInfo {
    fn id(&self) -> String;
    fn validate(&self, hash: &str) -> bool;
    fn requires_auth(&self) -> bool;
}

impl ApiKeyInfo for ApiKey {
    fn id(&self) -> String {
        self.id.to_string()
    }

    fn validate(&self, hash: &str) -> bool {
        let mut sha512 = Sha512::new();
        sha512.update(hash);

        format!("{:X}", sha512.finalize()) == self.hash
    }

    fn requires_auth(&self) -> bool {
        self.auth_required
    }
}

#[derive(Debug, Clone, Hash, PartialEq, Eq)]
pub(crate) struct ApiKey {
    pub(crate) id: String,
    pub(crate) key: Option<String>,
    pub(crate) hash: String,
    pub(crate) name: String,
    pub(crate) auth_required: bool,
}

impl ApiKey {
    pub(crate) fn create(name: &str, requires_auth: bool) -> Self {
        let uuid = Uuid::new_v4().simple();
        let key: String = create_random(KEY_LENGTH);
        let mut sha512 = Sha512::new();
        sha512.update(key.clone());

        ApiKey {
            id: uuid.to_string(),
            key: Some(key.to_owned()),
            hash: format!("{:X}", sha512.finalize()),
            auth_required: requires_auth,
            name: name.to_string(),
        }
    }

    pub(crate) fn validate(&self, password: &str) -> bool {
        let mut sha512 = Sha512::new();
        sha512.update(password);

        format!("{:X}", sha512.finalize()) == self.hash
    }
}

pub(crate) fn create_random(len: usize) -> String {
    rand::thread_rng()
        .sample_iter(&Alphanumeric)
        .take(len)
        .map(char::from)
        .collect()
}
