/// Login predicate middleware.
///
/// Requires that the user is authenticated.
#[macro_export]
macro_rules! login_required {
    ($backend_type:ty) => {{
        async fn is_authenticated(
            auth_session: $crate::axum_test::AuthenticationBackend<$backend_type>,
        ) -> bool {
            auth_session.is_authenticated()
        }

        $crate::predicate_required!(is_authenticated, axum::http::StatusCode::UNAUTHORIZED)
    }};
}

/// Permission predicate middleware.
///
/// Requires that the specified permissions, either user or group or both, are
/// all assigned to the user.
#[macro_export]
macro_rules! permission_required {
    ($backend_type:ty, login_url = $login_url:expr, redirect_field = $redirect_field:expr, $($perm:expr),+ $(,)?) => {{
        use $crate::AuthzBackend;

        async fn is_authorized(auth_session: $crate::axum_test::AuthSession<$backend_type>) -> bool {
            if let Some(ref user) = auth_session.user {
                let mut has_all_permissions = true;
                $(
                    has_all_permissions = has_all_permissions &&
                        auth_session.backend.has_perm(user, $perm.into()).await.unwrap_or(false);
                )+
                has_all_permissions
            } else {
                false
            }
        }

        $crate::predicate_required!(
            is_authorized,
            login_url = $login_url,
            redirect_field = $redirect_field
        )
    }};
}

/// Predicate middleware.
///
/// Can be specified with a login URL and next redirect field or an alternative
/// which implements [`IntoResponse`](axum::response::IntoResponse).
///
/// When the predicate passes, the request processes normally. On failure,
/// either a redirect to the specified login URL is issued or the alternative is
/// used as the response.
#[macro_export]
macro_rules! predicate_required {
    ($predicate:expr, $alternative:expr) => {{
        use axum::{
            extract::Request,
            middleware::{from_fn, Next},
            response::IntoResponse,
        };

        from_fn(
            |auth_session: $crate::axum_test::AuthenticationBackend<_>,
             req: Request,
             next: Next| async move {
                if $predicate(auth_session).await {
                    next.run(req).await
                } else {
                    $alternative.into_response()
                }
            },
        )
    }};
}
