use std::fmt::{self, Display};

use axum::{
    http::StatusCode,
    response::{IntoResponse, Response},
    Json,
};
use error_stack::Context;
use minisign::PError;
use serde::Serialize;
use sha2::digest::InvalidLength;
use tracing::error;

use crate::api::{backend::ApiBackend, routes::users::models::User};

#[derive(Debug)]
pub struct AppError;

impl fmt::Display for AppError {
    fn fmt(&self, fmt: &mut fmt::Formatter<'_>) -> fmt::Result {
        fmt.write_str("Error on executing Application")
    }
}

impl Context for AppError {}

#[derive(Debug, Serialize)]
pub struct ErrorInfo<'a> {
    error_message: &'a str,
    debug_info: Option<&'a String>,
}

#[derive(Debug, PartialEq)]
pub enum ApiError {
    SQLQueryError(String),
    InvalidCredentials,
    InternalError(String),
    AccessDenied,
}

impl ApiError {
    fn as_error_info(&self) -> (StatusCode, ErrorInfo) {
        let (status, error_message, debug_info) = match self {
            Self::InvalidCredentials => (StatusCode::UNAUTHORIZED, "Invalid credentials", None),
            Self::SQLQueryError(error) => (
                StatusCode::INTERNAL_SERVER_ERROR,
                "Invalid credentials",
                Some(error),
            ),
            Self::InternalError(error) => (
                StatusCode::INTERNAL_SERVER_ERROR,
                "Internal Server Error",
                Some(error),
            ),
            Self::AccessDenied => (StatusCode::FORBIDDEN, "Access Denied", None), // ApiError::WrongCredentials => (StatusCode::UNAUTHORIZED, "Wrong credentials"),
                                                                                  // ApiError::MissingCredentials => (StatusCode::BAD_REQUEST, "Missing credentials"),
                                                                                  // ApiError::TokenCreation => (StatusCode::INTERNAL_SERVER_ERROR, "Token creation error"),
                                                                                  // ApiError::InvalidToken => (StatusCode::BAD_REQUEST, "Invalid token"),
                                                                                  // ApiError::InvalidApiKey => (StatusCode::BAD_REQUEST, "Invalid api key"),
                                                                                  // ApiError::InternalServerError => {
                                                                                  //     (StatusCode::INTERNAL_SERVER_ERROR, "Unspecified error")
                                                                                  // }
                                                                                  // ApiError::AccessDenied => (StatusCode::UNAUTHORIZED, "Access denied"),
                                                                                  // ApiError::InvalidPermissions => (StatusCode::BAD_REQUEST, "Invalid permissions"),
        };

        (
            status,
            ErrorInfo {
                error_message,
                debug_info,
            },
        )
    }
}

impl IntoResponse for ApiError {
    fn into_response(self) -> Response {
        let (status, error_info) = self.as_error_info();
        let body = Json(error_info);
        (status, body).into_response()
    }
}

impl From<sqlx::Error> for ApiError {
    fn from(error: sqlx::Error) -> Self {
        error!("{error}");
        Self::SQLQueryError(error.to_string())
    }
}

impl From<axum_jwt_login::Error<User, ApiBackend>> for ApiError {
    fn from(value: axum_jwt_login::Error<User, ApiBackend>) -> Self {
        match value {
            axum_jwt_login::Error::Jwt(error) => ApiError::InternalError(error.to_string()),
            axum_jwt_login::Error::Backend(error) => error,
            _ => unreachable!(),
        }
    }
}

impl From<argon2::Error> for ApiError {
    fn from(_: argon2::Error) -> Self {
        Self::InvalidCredentials
    }
}

impl From<InvalidLength> for ApiError {
    fn from(value: InvalidLength) -> Self {
        Self::InternalError(format!("Invalid HMac Key length: {value}"))
    }
}

impl From<PError> for ApiError {
    fn from(value: PError) -> Self {
        Self::InternalError(format!("MiniSign Error: {value}"))
    }
}

impl From<serde_json::Error> for ApiError {
    fn from(value: serde_json::Error) -> Self {
        Self::InternalError(format!("Error on (de)serialization: {value}"))
    }
}

impl Display for ApiError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let error_info = self.as_error_info().1;
        f.write_str(&format!(
            "{}: {}",
            error_info.error_message,
            error_info.debug_info.unwrap_or(&"".to_string())
        ))
    }
}

impl std::error::Error for ApiError {}
