use std::{path::PathBuf, time::Duration};

use clap::Parser;
use error_stack::{Report, ResultExt};
use errors::AppError;
use once_cell::sync::Lazy;
use tokio::signal;
use tokio_util::task::TaskTracker;
use tracing::{error, info, level_filters::LevelFilter, Level};
use tracing_rolling_file::{RollingConditionBase, RollingFileAppenderBase};
use tracing_subscriber::{fmt, layer::SubscriberExt, util::SubscriberInitExt};

pub mod api;
mod cli;
mod config;
mod errors;
mod utils;

pub static ROOT_PATH: Lazy<PathBuf> = Lazy::new(|| match std::env::current_exe() {
    #[cfg(not(debug_assertions))]
    Ok(path) => path,
    #[cfg(debug_assertions)]
    Ok(_) => PathBuf::from("./"),
    Err(e) => {
        error!("Exiting on error for getting root path: {e:?}");
        panic!("Exiting on error for getting root path: {e:?}");
    }
});

pub const APP_NAME: &str = "Generic API Service";

#[tokio::main]
async fn main() -> Result<(), Report<AppError>> {
    let run_exit_code = run().await;

    if let Err(e) = run_exit_code.as_ref() {
        error!("{e}");
        tokio::time::sleep(Duration::from_secs(1)).await;
    }

    run_exit_code
}

async fn run() -> Result<(), Report<AppError>> {
    dotenv::dotenv().ok();

    // Set Report Colour Mode to NONE
    Report::set_color_mode(error_stack::fmt::ColorMode::None);

    // prepare CLI
    let cli = cli::Cli::parse();

    // load config file
    let mut config = config::Configuration::new(cli.config())?;
    config.check()?;
    println!("{config:?}");

    let level_filter = LevelFilter::from_level(match config.debug {
        true => Level::DEBUG,
        false => Level::INFO,
    });

    // Prepare logging to file
    let file_appender = RollingFileAppenderBase::new(
        ROOT_PATH.with_file_name(format!("{}.log", env!("CARGO_PKG_NAME"))),
        RollingConditionBase::new().max_size(1024 * 1024 * 2),
        5,
    )
    .change_context(AppError)?;
    let (non_blocking, _guard) = tracing_appender::non_blocking(file_appender);

    // prepare initialization of logging
    let log_layers = tracing_subscriber::registry().with(level_filter).with(
        fmt::Layer::default()
            .with_target(false)
            .with_ansi(false)
            .with_writer(non_blocking),
    );

    // also log to console in debug mode
    #[cfg(debug_assertions)]
    let stdout_log = tracing_subscriber::fmt::layer().pretty();
    #[cfg(debug_assertions)]
    let log_layers = log_layers.with(stdout_log);

    // Initialize logging
    log_layers.init();

    // handle CLI input
    let daemon = cli.handle(&mut config).await?;

    match daemon {
        cli::DaemonStatus::NotRunning => {}
        cli::DaemonStatus::Running((cancellation_token, shutdown_signal)) => {
            // create task tracker
            let tracker = TaskTracker::new();

            // terminate signal
            #[cfg(unix)]
            let terminate = async {
                signal::unix::signal(signal::unix::SignalKind::terminate())
                    .expect("Failed to install termination signal handler")
                    .recv()
                    .await;
            };
            #[cfg(not(unix))]
            let terminate = std::future::pending::<()>();

            let shutdown_signal = async {
                match shutdown_signal {
                    Some(mut signal) => signal.recv().await,
                    None => std::future::pending::<Option<()>>().await,
                };
            };

            // wait for shutdown signals
            tokio::select! {
                _ = signal::ctrl_c() => {},
                _ = terminate => {},
                _ = shutdown_signal => {},
            }

            // send shutdown signal to application and wait
            cancellation_token.cancel();

            // close task tracker
            tracker.close();

            // Wait for everything to finish.
            tracker.wait().await;
            tokio::time::sleep(Duration::from_millis(200)).await;

            info!("{APP_NAME} gracefully shut down.");
        }
    };

    // let _ = axum_test::start_axum().await.unwrap();

    // let _ = ldap_test::try_authentication().await.unwrap();

    Ok(())
}
